# Filename:     Controller.py
# Written by:   Niranjan Bhujel
# Description:  Contains controller such as LQR, MPC, etc.


from math import inf, isinf
from pynlcontrol.BasicUtils import Integrate
import casadi as ca


def LQR(A, B, C, D, Q, R, Qt, Ts, horizon=inf, reftrack=False, NMAX=1000, tol=1e-5, Integrator='rk4'):
    """
    Function to implement discrete-time linear quadratic regulator (LQR).

    Parameters
    ----------
    A : numpy.2darray or casadi.SX.array
        Continuous time state matrix
    B : numpy.2darray or casadi.SX.array
        Continuous time input matrix
    C : numpy.2darray or casadi.SX.array
        Continuous time output matrix
    D : numpy.2darray or casadi.SX.array
        Continuous time output matrix coefficient of input
    Q : numpy.2darray or casadi.SX.array
        Weight to penalize control error
    R : numpy.2darray or casadi.SX.array
        Weight to penalize control effort
    Qt : numpy.2darray or casadi.SX.array
        Weight of terminal cost to penalize control error
    Ts : float
        Sample time of controller
    horizon : int, optional
        Horizon length of LQR. Defaults to inf for infinite horizon LQR problem.
    reftrack : bool, optional
        Whether problem is reference tracking. Defaults to False.
    NMAX : int, optional
        Maximum iteration for solving matrix Ricatti equation. Defaults to 1000.
    tol : float, optional
        Tolerance for solution of matrix Ricatti equation. Defaults to 1e-5.
    Integrator : str, optional
        Integrator to be used for discretization. Defaults to 'rk4'.

    Returns
    -------
    tuple
        Tuple of Input, Output, Input name and Output name. Inputs are x or [x, r] (depending upon the problem is reference tracking or not) and output are u and K.
            Input and output are casadi symbolics (`casadi.SX`).
            These inputs are and outputs can be mapped using `casadi.Function` which can further be code generated.
    
    Example
    -------
    >>> from pynlcontrol import Controller, BasicUtils
    >>> import casadi as ca
    >>> Q11 = ca.SX.sym('Q11')
    >>> Q22 = ca.SX.sym('Q22')
    >>> Q33 = ca.SX.sym('Q33')
    >>> Q = BasicUtils.directSum([Q11, Q22, Q33])
    >>> R11 = ca.SX.sym('R11')
    >>> R22 = ca.SX.sym('R22')
    >>> R = BasicUtils.directSum([R11, R22])
    >>> A = ca.SX([[-0.4,0.1,-2],[0,-0.3,4],[1,0,0]])
    >>> B = ca.SX([[1,1],[0,1],[1,0]])
    >>> C = ca.SX([[1, 0, 0], [0, 1, 0]])
    >>> D = ca.SX([[0, 0], [0, 0]])
    >>> In, Out, InName, OutName = Controller.LQR(A=A, B=B, C=C, D=D, Q=Q, R=R, Qt=Q, Ts=0.1, horizon=10, reftrack=True)
    >>> lqr_func = ca.Function('lqr_func', In + [Q11, Q22, Q33, R11, R22], Out, InName + ['Q11', 'Q22', 'Q33', 'R11', 'R22'], OutName)
    >>> BasicUtils.Gen_Code(lqr_func, 'lqr_code', printhelp=True)
    x(3, 1), ref(2, 1), Q11(1, 1), Q22(1, 1), Q33(1, 1), R11(1, 1), R22(1, 1) -> u(2, 1), K(6, 1)
    lqr_code.c
    lqr_code_Call.c
    #include "lqr_code.h"
    #include "lqr_code_Call.h"
    lqr_code_Call_Func(x, ref, Q11, Q22, Q33, R11, R22, u, K);


    Running above code generates C-codes for LQR implementation. Implementation using Simulink can be found in example folder.
    """

    nX = A.shape[0]
    nU = B.shape[1]
    nY = C.shape[0]

    assert D.shape[1] == B.shape[1], "Inconsistent shape of B and D"
    assert nY == nU, "Number of control inputs and controlled outputs should be same."
    assert nX==Q.shape[0]==Q.shape[1], "Error in size of Q"
    assert nX==Qt.shape[0]==Qt.shape[1], "Error in size of Qt"
    assert nY==R.shape[0]==R.shape[1], "Error in size of R"

    x = ca.SX.sym('x', nX, 1)
    u = ca.SX.sym('u', nU, 1)

    def Fc(x, u):
        return A @ x + B @ u

    xk1 = Integrate(Fc, Integrator, Ts, x, u)
    Ad = ca.jacobian(xk1, x)
    Bd = ca.jacobian(xk1, u)

    P = Qt
    MAXITER = NMAX if isinf(horizon) else horizon
    for _ in range(MAXITER):
        P = Ad.T @ P @ Ad - (Ad.T @ P @ Bd) @ ca.inv(R +
                                                     Bd.T @ P @ Bd) @ (Bd.T @ P @ Ad) + Q

    K = ca.inv(R + Bd.T @ P @ Bd) @ (Bd.T @ P @ Ad)

    u = -K @ x

    if reftrack:
        r = ca.SX.sym('r', nY, 1)
        tmp = ca.vertcat(
            ca.horzcat(A, B),
            ca.horzcat(C, D)
        )

        Nxu = ca.inv(tmp) @ ca.vertcat(ca.GenSX_zeros(nX, nY), ca.SX_eye(nY))
        Nx = Nxu[0:nX, :]
        Nu = Nxu[nX:,:]

        u += (Nu + K@Nx)@r

        return [x, r], [u, K.T.reshape((-1, 1))], ['x', 'ref'], ['u', 'K']
    else:
        return [x], [u, K.T.reshape((-1, 1))], ['x'], ['u', 'K']



# Filename:     BasicUtils.py
# Written by:   Niranjan Bhujel
# Description:  Contains basic functions such as generating interface code for casadi function,
#               numerical integration, nlp to generalized Gauss Newton (GGN), etc.


import casadi as ca
import shutil
import os
import sys
from jinja2 import Environment, FileSystemLoader


def Gen_Code(func, filename, dir='/', mex=False, printhelp=False, optim=False):
    """
    Function to generate c code (casadi generated as well as interface) for casadi function.

    Generates c and h file (from CasADi) as well as interface code. Using interface code, the generated
    codes can be integrated with other platform such as Simulink, PSIM, etc.

    Parameters
    ----------
    func : casadi.Function
        CasADi function for which code needs to be generated. This function maps input to output.
    filename : str
        File name of the generated code. Note: Filename should not contain "_Call" in it.
    dir : str, optional
        Directory where codes need to be generated. Defaults to current directory.
    mex : bool, optional
        Option if mex is required. Defaults to False.
    printhelp : bool, optional
        Option if information about input/output and its size are to be printed . If mex is False, sfunction help is also printed. Defaults to False.
    optim : bool, optional
        Whether code is being generated for CasADi optimization problem. Defaults to False.

    See Also
    --------
    Gen_Test: Generates main function code which can be compiled and debugged.

    Example
    --------
    >>> import casadi as ca
    >>> import numpy
    >>> x = ca.SX.sym('x')
    >>> y = ca.SX.sym('y', 2)
    >>> f1 = x + y[0]
    >>> f2 = x + y[1]**2
    >>> Func = ca.Function(
        'Func',
        [x, y],
        [f1, f2],
        ['x', 'y'],
        ['f1', 'f2'],
        )
    >>> BasicUtils.Gen_Code(Func, 'GenCodeTest', dir='./', mex=False, printhelp=True)
    x(1, 1), y(2, 1) -> f1(1, 1), f2(1, 1)
    GenCodeTest.c
    GenCodeTest_Call.c
    #include "GenCodeTest.h"
    #include "GenCodeTest_Call.h"
    GenCodeTest_Call_Func(x, y, f1, f2);


    Above code creates `GenCodeTest.c`, `GenCodeTest_Call.c`, `GenCodeTest.h` and `GenCodeTest_Call.h` that evaluates `f1` and `f2` from value of `x` and `y`. `x` and `y` in the above example should be declared as pointer.
    """

    funcName = func.name()

    if dir == '/' or dir == './':
        dir = ''
    else:
        if dir not in os.listdir():
            os.mkdir(dir)
        if dir[-1] != '/':
            dir += '/'

    In = [f"{x}" for x in func.name_in()]
    Out = [f"{x}" for x in func.name_out()]

    if mex:
        func.generate(f'{filename}.c', {'main': False,
                                        'mex': True, 'with_header': True})
        shutil.move(f'{filename}.c', f'{dir}{filename}.c')
        shutil.move(f'{filename}.h', f'{dir}{filename}.h')
    else:
        func.generate(f'{filename}.c', {'main': False,
                                        'mex': False, 'with_header': True})
        shutil.move(f'{filename}.c', f'{dir}{filename}.c')
        shutil.move(f'{filename}.h', f'{dir}{filename}.h')

        SizeIn = []
        for k in range(len(In)):
            SizeIn.append(func.size_in(k)[0]*func.size_in(k)[1])

        SizeOut = []
        for k in range(len(Out)):
            SizeOut.append(func.size_out(k)[0]*func.size_out(k)[1])

        totalIOsize = 0
        for k in range(len(In)):
            totalIOsize += func.size_in(k)[0]*func.size_in(k)[1]
        for k in range(len(Out)):
            totalIOsize += func.size_out(k)[0]*func.size_out(k)[1]

        # template rendering
        templatePath = f"{os.path.dirname(os.path.realpath(__file__))}/templates"

        file_loader = FileSystemLoader(templatePath)
        env = Environment(loader=file_loader)

        # template for c file
        template = env.get_template('GenCode_c.j2')

        CodeOutput_c = template.render(
            FileName=filename,
            Func_Name=funcName,
            Func_In=In,
            Func_Out=Out,
            arg_size=func.sz_arg(),
            res_size=func.sz_res(),
            iw_size=func.sz_iw(),
            w_size=int(optim)*func.sz_w() + totalIOsize,
            SizeIn=SizeIn,
            SizeOut=SizeOut,
        )

        # template for h file
        template = env.get_template('GenCode_h.j2')

        CodeOutput_h = template.render(
            FileName=filename,
            Func_Name=funcName,
            Func_In=In,
            Func_Out=Out,
            arg_size=func.sz_arg(),
            res_size=func.sz_res(),
            iw_size=func.sz_iw(),
            w_size=func.sz_w(),
            SizeIn=SizeIn,
            SizeOut=SizeOut,
        )

        # Write code to file
        with open(f"{dir}{filename}_Call.c", 'w') as fw:
            fw.write(CodeOutput_c)

        with open(f"{dir}{filename}_Call.h", 'w') as fw:
            fw.write(CodeOutput_h)

    # print information about input/out and its size.
    if printhelp:
        tmpIn = []
        for k in range(len(In)):
            tmpIn.append(
                f"{In[k]}({func.size_in(k)[0]}, {func.size_in(k)[1]})")

        tmpOut = []
        for k in range(len(Out)):
            tmpOut.append(
                f"{Out[k]}({func.size_out(k)[0]}, {func.size_out(k)[1]})")
        print(", ".join(tmpIn) + " -> " + ", ".join(tmpOut))
        print("\n")

        if ~mex:
            print(f"{dir}{filename}.c\n{dir}{filename}_Call.c\n\n")
            print(
                f'#include "{dir}{filename}.h"\n#include "{dir}{filename}_Call.h"\n\n')
            print(f'{filename}_Call_Func({", ".join(In)}, {", ".join(Out)});\n')


def Gen_Test(headers, varsIn, sizeIn, varsOut, sizeOut, callFuncName, filename, dir='/'):
    """
    Generates C code with main function. This code along with code generated by `Gen_Code` function can be compiled to executable to check computation time or debug.
    It can also be used as example to compile on other target.

    Parameters
    ----------
    headers : list[str]
        List of header files name along with extension `.h`. These files will be added in main file as #include "header".
    varsIn : list[str]
        List of name of input variables.
    sizeIn : list[int]
        List of size of input variables.
    varsOut : list(str)
        List of name of output variables.
    sizeOut : list[int]
        List of size of output variables.
    callFuncName : list(str)
        Name of the C function that needs to be called.
    filename : str
        Name of the generated C file (along with extension .c).
    dir : str, optional
        Directory where code needs to be generated. Defaults to current directory.

    Example
    _______
    >>> from pynlcontrol import BasicUtils
    >>> import casadi as ca
    >>> x = ca.SX.sym('x')
    >>> y = ca.SX.sym('y', 2)
    >>> f1 = x + y[0]
    >>> f2 = x + y[1]**2
    >>> Func = ca.Function(
        'Func',
        [x, y],
        [f1, f2],
        ['x', 'y'],
        ['f1', 'f2'],
        )
    >>> BasicUtils.Gen_Code(Func, 'GenCodeTest', dir='./', mex=False, printhelp=False)
    >>> BasicUtils.Gen_Test(
        headers=['GenCodeTest.h', 'GenCodeTest_Call.h'],
        varsIn=['x', 'y'],
        sizeIn=[1, 2],
        varsOut=['f1', 'f2'],
        sizeOut=[1, 1],
        callFuncName='GenCodeTest_Call_Func',
        filename='MainTest.c',
        dir='./',
        )

    
    All the line except last one is same as from `Gen_Code()` function. The last line generates filename with `MainTest.c` which can be used to test and debug the generated code.
    """



    if dir == '/' or dir == './':
        dir = ''
    else:
        if dir not in os.listdir():
            os.mkdir(dir)
        if dir[-1] != '/':
            dir += '/'

    templatePath = f"{os.path.dirname(os.path.realpath(__file__))}/templates"

    file_loader = FileSystemLoader(templatePath)
    env = Environment(loader=file_loader)

    # template for c file
    template = env.get_template('TestCode.j2')

    CodeOutput = template.render(
        headers=headers,
        sizeIn=sizeIn,
        varsIn=varsIn,
        sizeOut=sizeOut,
        varsOut=varsOut,
        callFuncName=callFuncName
    )

    with open(f"{dir}{filename}", 'w') as fw:
        fw.write(CodeOutput)


def Integrate(odefun, method, Ts, x0, u0, *args):
    """
    Function to integrate continuous-time ODE. It discretize the provided ODE function and gives value of state variables at next discrete time. 

    Parameters
    ----------
    odefun : function
        Python function with states as first and control input as second argument. Remaining argument could be anything required by state equations.
    method : str
        Method to integrate ODE. Supported methods are: 'FEuler', 'rk2', 'rk3', 'ssprk3', 'rk4', 'dormandprince'.
    Ts : float
        Step time to solve ODE
    x0 : float or casadi.SX or numpy.1darray
        Current states of the system
    u0 : float or casadi.SX or numpy.1darray
        Current control input to the system

    Returns
    -------
    float or casadi.SX or numpy.1darray
        Next states of the system.

    Example
    -------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> def Fc(x, u):
        x1 = x[0]
        x2 = x[1]
        return np.array([(1-x2**2)*x1 - x2 + u, x1])
    >>> T = 10
    >>> Ts = 0.1
    >>> t = np.arange(0, T+Ts, Ts)
    >>> x = np.zeros((2, t.shape[0]))
    >>> for k in range(t.shape[0]-1):
        u = 0 if t[k] < 1 else 1.0
        x[:,k+1] = Integrate(Fc, 'rk4', Ts, x[:,k], u)
    >>> plt.plot(t, x[0,:])
    >>> plt.plot(t, x[1,:])
    >>> plt.xlabel('time (s)')
    >>> plt.ylabel('$x_1$ and $x_2$')
    >>> plt.show()

    The above code integrates the ODE defined by function Fc from 0 to 10 s with step-time of 0.1 s using RK4 method.
    """
    if method == 'FEuler':
        xf = x0 + Ts*odefun(x0, u0, *args)
    elif method == 'rk2':
        k1 = odefun(x0, u0, *args)
        k2 = odefun(x0 + Ts/2*k1, u0, *args)
        xf = x0 + Ts*k2
    elif method == 'rk3':
        k1 = odefun(x0, u0, *args)
        k2 = odefun(x0+Ts/2*k1, u0, *args)
        k3 = odefun(x0+Ts*k2, u0, *args)
        xf = x0 + Ts/6*(k1 + 4*k2 + k3)
    elif method == 'rk4':
        k1 = odefun(x0, u0, *args)
        k2 = odefun(x0+Ts/2*k1, u0, *args)
        k3 = odefun(x0+Ts/2*k2, u0, *args)
        k4 = odefun(x0 + Ts*k3, u0, *args)
        xf = x0 + Ts/6*(k1 + 2*k2 + 2*k3 + k4)
    elif method == 'ssprk3':
        k1 = odefun(x0, u0, *args)
        k2 = odefun(x0+Ts*k1, u0, *args)
        k3 = odefun(x0+Ts/2*k2, u0, *args)
        xf = x0 + Ts/6*(k1 + k2 + 4*k3)
    elif method == 'dormandprince':
        k1 = odefun(x0, u0, *args)
        k2 = odefun(x0 + 1/5*Ts*k1, u0, *args)
        k3 = odefun(x0 + 3/10*Ts*k2, u0, *args)
        k4 = odefun(x0 + 4/5*Ts*k3, u0, *args)
        k5 = odefun(x0 + 8/9*Ts*k4, u0, *args)
        k6 = odefun(x0 + 1*Ts*k5, u0, *args)
        k7 = odefun(x0 + 1*Ts*k6, u0, *args)
        # Xf = x0 + Ts*(35/384*k1+500/1113*k3+125/192*k4-2187/6784*k5+11/84*k6)
        xf = x0 + Ts*(5179/57600*k1 + 7571/16695*k3 + 393 /
                      640*k4 - 92097/339200*k5 + 187/2100*k6 + 1/40*k7)
    else:
        print(f'Error: No integrator method {method} found.')
        sys.exit(1)
    return xf


def nlp2GGN(z, J, g, lbg, ubg, p):
    """
    Converts provided nonlinear programming into quadratic form using generalized Gauss-Newton method.

    Parameters
    ----------
    z : casadi.SX
        Vector of unknown variables of optimization problem
    J : casadi.SX
        Object function of the optimization problem
    g : casadi.SX
        Vector of constraints function
    lbg : casadi.SX
        Vector of lower limits on constraint function
    ubg : casadi.SX
        Vector of upper limits on constraint function
    p : casadi.SX
        Vector of input to the optimization problem

    Returns
    -------
    dict
        Dictionary of optimization problem.
        keywords
            x: Vector of decision variables 
            f: New quadratic cost function
            g: New constraint function
            lbg: Lower limits on constraint function
            ubg: Upper limits on constraint function
    """
    zOp = ca.SX.sym('zOp', z.shape)

    Jnew = ca.substitute(J, z, zOp) + \
        ca.substitute(ca.jacobian(J, z), z, zOp) @ (z - zOp)

    Cost = ca.norm_2(Jnew)**2

    nlp = {
        'x': z,
        'f': Cost
    }

    if g is not None:
        gnew = ca.substitute(g, z, zOp) + \
            ca.substitute(ca.jacobian(g, z), z, zOp) @ (z - zOp)
        nlp['g'] = gnew
        nlp['lbg'] = lbg
        nlp['ubg'] = ubg

    if p is not None:
        nlp['p'] = p

    nlp['zOp'] = zOp

    return nlp


def __SXname__(x):
    """
    Returns the name of casadi.SX symbolics

    Parameters
    ----------
    x : list[casadi.SX] or casadi.SX
        List of casadi symbolics or just casadi symbolics.

    Returns
    -------
    list[str] or str
        List of name of symbolics or just name of symbolics
    """
    if isinstance(x, list):
        nameList = []
        for tmp in x:
            z = str(tmp)
            z = z.split('_')
            nameList.append(z[0])

        return nameList
    else:
        z = str(x[0])
        z = z.split('_')
        return z[0]


def directSum(A):
    """
    Direct sum of matrices in the list A.

    Parameters
    ----------
    A : list
        List of matrices.

    Returns
    -------
    casadi.SX.sym
        Direct sum of all matrices in list A.

    Example
    -------
    >>> import casadi as ca
    >>> from pynlcontrol import BasicUtils
    >>> import casadi as ca
    >>> from pynlcontrol import BasicUtils
    >>> A1 = ca.SX.sym('A1', 2, 2)
    >>> A2 = ca.SX.sym('A2', 3, 3)
    >>> A = [A1, A2]
    >>> BasicUtils.directSum(A)
    SX(@1=0,
    [[A1_0, A1_2, @1, @1, @1],
     [A1_1, A1_3, @1, @1, @1],
     [@1, @1, A2_0, A2_3, A2_6],
     [@1, @1, A2_1, A2_4, A2_7],
     [@1, @1, A2_2, A2_5, A2_8]])

    Above code puts matrices `A1` and `A2` in the diagonal and fills zero elsewhere.
    """

    assert isinstance(A, list), "Argument must be list."
    N = len(A)
    NRow = 0
    NCol = 0
    for i in range(N):
        NRow += A[i].shape[0]
        NCol += A[i].shape[1]
    z = ca.GenSX_zeros((NRow, NCol))
    Row_Idx = 0
    Col_Idx = 0
    for i in range(N):
        z[Row_Idx:Row_Idx+A[i].shape[0], Col_Idx:Col_Idx+A[i].shape[1]] = A[i]
        Row_Idx += A[i].shape[0]
        Col_Idx += A[i].shape[1]
    return z


# Filename:     Estimator.py
# Written by:   Niranjan Bhujel
# Description:  Contains estimators such as kalman filter, extended kalman filter and simple moving horizon estimators


from pynlcontrol.BasicUtils import Integrate, nlp2GGN
import casadi as ca


def KF(A, B, C, D, Qw, Rv, Ts, Integrator='rk4'):
    """
    Function to implement Kalman filter (KF). 

    Parameters
    ----------
    A: (numpy.2darray or casadi.SX array)
        Continuous-time state matrix of the system
    B: (numpy.2darray or casadi.SX array)
        Continuous-time input matrix of the system
    C: (numpy.2darray or casadi.SX array)
        Continuous-time measurement matrix of the system
    D: (numpy.2darray or casadi.SX array)
        Continuous time output matrix coefficient of input
    Qw: (numpy.2darray or casadi.SX array)
        Process noise covariance matrix
    Rv: (numpy.2darray or casadi.SX array)
        Measurement noise covariance matrix
    Ts: (float)
        Sample time of KF
    Integrator: (str, optional)
        Integrator to be used for discretization. Defaults to 'rk4'.

    Returns
    -------
    tuple
        Tuple of Input, Output, Input name and Output name. Inputs are u, y, xp, Pp and output are xhat and Phat. Input and output are casadi symbolics (`casadi.SX`).
            u: Current input to the system
            y: Current measurement of the system
            xp: State estimate from previous discrete time
            Pp: Covariance estimate from previous discrete time (reshaped to column matrix)
            xhat: State estimate at current discrete time
            Phat: Covariance estimate at current discrete time (reshaped to column matrix)

            These inputs are and outputs can be mapped using `casadi.Function` which can further be code generated.

    Example
    -------
    >>> from pynlcontrol import Estimator, BasicUtils
    >>> import casadi as ca
    >>> Q11 = ca.SX.sym('Q11')
    >>> Q22 = ca.SX.sym('Q22')
    >>> Q33 = ca.SX.sym('Q33')
    >>> Q = BasicUtils.directSum([Q11, Q22, Q33])
    >>> R22 = ca.SX.sym('R22')
    >>> R = BasicUtils.directSum([R11, R22])
    >>> A = ca.SX([[-0.4,0.1,-2],[0,-0.3,4],[1,0,0]])
    >>> B = ca.SX([[1,1],[0,1],[1,0]])
    >>> C = ca.SX([[1, 0, 0], [0, 1, 0]])
    >>> D = ca.SX([[0, 0], [0, 0]])
    >>> In, Out, InName, OutName = Estimator.KF(A, B, C, D, Q, R, 0.1)
    >>> KF_func = ca.Function('KF_func', In + [Q11, Q22, Q33, R11, R22], Out, InName + ['Q11', 'Q22', 'Q33', 'R11', 'R22'], OutName)
    >>> BasicUtils.Gen_Code(KF_func, 'KF_code', printhelp=True)
    u(2, 1), y(2, 1), xhatp(3, 1), Pkp(9, 1), Q11(1, 1), Q22(1, 1), Q33(1, 1), R11(1, 1), R22(1, 1) -> xhat(3, 1), Phat(9, 1)
    KF_code.c
    KF_code_Call.c
    #include "KF_code.h"
    #include "KF_code_Call.h"
    KF_code_Call_Func(u, y, xhatp, Pkp, Q11, Q22, Q33, R11, R22, xhat, Phat);


    Running above code generates C-codes for KF implementation. Implementation using Simulink can be found in example folder.
    """
    nX = A.shape[0]
    nU = B.shape[1]
    nY = C.shape[0]

    x = ca.SX.sym('x', nX, 1)
    u = ca.SX.sym('u', nU, 1)

    def Fc(x, u):
        return A @ x + B @ u

    xk1 = Integrate(Fc, Integrator, Ts, x, u)
    Ad = ca.jacobian(xk1, x)
    Bd = ca.jacobian(xk1, u)
    Cd = C
    Dd = D

    xp = ca.SX.sym('xp', nX, 1)
    u = ca.SX.sym('u', nU, 1)
    y = ca.SX.sym('y', nY, 1)

    Pp = ca.SX.sym('Pp', nX*nX, 1)
    Ppt = ca.reshape(Pp, nX, nX)

    xkm = Ad @ xp + Bd @ u
    Pkm = Ad @ Ppt @ ca.transpose(Ad) + Qw
    yr = y - (Cd @ xkm + Dd @ u)
    Sk = Cd @ Pkm @ ca.transpose(Cd) + Rv

    Kk = Pkm @ ca.transpose(Cd) @ ca.inv(Sk)

    xhat = xkm + Kk @ yr
    Phat = (ca.SX_eye(Ad.shape[0]) - Kk @ Cd) @ Pkm

    return [u, y, xp, Pp], [xhat, ca.reshape(Phat, nX*nX, 1)], ['u', 'y', 'xhatp', 'Pkp'], ['xhat', 'Phat']


def EKF(nX, nU, nY, F, H, Qw, Rv, Ts, Integrator='rk4'):
    """
    Function to implement Extended Kalman filter (EKF).

    Parameters
    ----------
    nX: (int)
        Number of state variables
    nU: (int)
        Number of control inputs
    ny: (int)
        Number of measurement outputs
    F: (function)
        Function that returns right-hand side of state differential equation. Input arguments to F should be states and inputs respectively.
    H: (function)
        Function that retuns measurement variable from state and input variables. nput arguments to H should be states and inputs respectively.
    Qw: (numpy.2darray or casadi.SX array)
        Process noise covariance matrix
    Rv: (numpy.2darray or casadi.SX array)
        Measurement noise covariance matrix
    Ts: (float)
        Sample time of the Kalman filter.
    Integrator: (str, optional)
        Integration method. Defaults to 'rk4'. For list of supported integrator, please see documentation of function `Integrate()`.

    Returns
    -------
    tuple: 
        Tuple of Input, Output, Input name and Output name. Inputs are u, y, xp, Pp and output are xhat and Phat. Input and output are casadi symbolics (`casadi.SX`).
            u: Current input to the system
            y: Current measurement of the system
            xp: State estimate from previous discrete time
            Pp: Covariance estimate from previous discrete time (reshaped to column matrix)
            xhat: State estimate at current discrete time
            Phat: Covariance estimate at current discrete time (reshaped to column matrix)

            These inputs are and outputs can be mapped using `casadi.Function` which can further be code generated.
    """
    assert isinstance(nX, int), "nX must be integer."
    assert isinstance(nU, int), "nU must be integer."
    assert isinstance(nY, int), "nY must be integer."

    assert Qw.shape[0] == Qw.shape[1], "Qw is not square matrix."
    assert Rv.shape[0] == Rv.shape[1], "Rv is not square matrix."

    assert nX == Qw.shape[0], "Shape mismatch of Qw with nX."

    assert nY == Rv.shape[0], "Shape mismatch of Rv with nY."

    assert isinstance(Ts, float), "Sample time (Ts) must be float."

    xp = ca.SX.sym('xp', nX, 1)
    u = ca.SX.sym('u', nU, 1)
    y = ca.SX.sym('y', nY, 1)

    Pp = ca.SX.sym('Pp', nX*nX, 1)
    Ppt = ca.reshape(Pp, nX, nX)

    xkm = Integrate(F, Integrator, Ts, xp, u)
    Fk = ca.jacobian(xkm, xp)

    Pkm = Fk @ Ppt @ ca.transpose(Fk) + Qw
    yr = y - H(xkm, u)

    tmpfun = ca.Function('tmpfun', [xp], [ca.jacobian(H(xp), xp)])
    Hk = tmpfun(xkm)

    Sk = Hk @ Pkm @ ca.transpose(Hk) + Rv
    Kk = Pkm @ ca.transpose(Hk) @ ca.inv(Sk)

    xhat = xkm + Kk @ yr
    Phat = (ca.SX_eye(Fk.shape[0]) - Kk @ Hk) @ Pkm

    return [u, y, xp, Pp], [xhat, ca.reshape(Phat, nX*nX, 1)], ['u', 'y', 'xhatp', 'Pkp'], ['xhat', 'Phat']


def UKF(nX, nU, nY, F, H, Qw, Rv, Ts, PCoeff=None, Wm=None, Wc=None, alpha=1.0e-3, beta=2.0, kappa=0.0, Integrator='rk4'):
    """
    Function to implement Unscented Kalman filter (UKF). 
    
    If either of PCoeff or Wm or Wc is None, it calculates those values with alpha=1e-3, Beta=2 and kappa=0. To use manual weights, specify PCOeff, Wm and Wc. Otherwise, use alpha, beta and kappa parameters to set those values.

    Parameters
    ----------
    nX: (int)
        Number of state variables
    nU: (int)
        Number of control inputs
    nY: (int)
        Number of measurement outputs
    F: (function)
        Function that returns right-hand side of state differential equation. Input arguments to F should be states and inputs respectively.
    H: (function)
        Function that retuns measurement variable from state and input variables. nput arguments to H should be states and inputs respectively.
    Qw: (numpy.2darray or casadi.SX array)
        Process noise covariance matrix
    Rv: (numpy.2darray or casadi.SX array)
        Measurement noise covariance matrix
    Ts: (float)
        Sample time of the Kalman filter.
    PCoeff: (float)
        Coefficient of covariance matrix (inside square root term) when calculating sigma points. Defaults to None
    Wm: (list, optional)
        List of weights for mean calculation. Defaults to None.
    Wc: (list, optional)
        List of weights for covariance calculation. Defaults to None.
    alpha: (float, optional)
        Value of alpha parameter. Defaults to 1.0e-3.
    beta: (float, optional)
        Value of beta parameter. Defaults to 2.0.
    kappa: (float, optional)
        Value of kappa parameter. Defaults to 0.0.
    Integrator: (str, optional)
        Integration method. Defaults to 'rk4'. For list of supported integrator, please see documentation of function `Integrate`.

    Returns
    -------
    tuple:
        Tuple of Input, Output, Input name and Output name. Inputs are u, y, xp, Pp and output are xhat and Phat. Input and output are casadi symbolics (`casadi.SX`).
            u: Current input to the system
            y: Current measurement of the system
            xp: State estimate from previous discrete time
            Pp: Covariance estimate from previous discrete time (reshaped to column matrix)
            xhat: State estimate at current discrete time
            Phat: Covariance estimate at current discrete time (reshaped to column matrix)

            These inputs are and outputs can be mapped using `casadi.Function` which can further be code generated.
    """
    def SigmaMean(S, Wm):
        N = len(Wm)
        SMean = ca.GenSX_zeros(S.shape[0], 1)
        for k in range(N):
            SMean += Wm[k] * S[:, [k]]
        return SMean


    def SigmaCovar(S1, S2, Wm, Wc):
        N = len(Wm)
        SCov = ca.GenSX_zeros(S1.shape[0], S2.shape[0])
        for k in range(N):
            SCov += Wc[k] * (S1[:, k] - SigmaMean(S1, Wm)) @ ca.transpose((S2[:, [k]] - SigmaMean(S2, Wm)))
        return SCov

    def GenSigma(xbar, P, PCoeff):
        nx = xbar.shape[0]
        S = ca.GenSX_zeros((nx, 2*nx+1))
        S[:, [0]] = xbar

        L = ca.chol(P).T

        for k in range(1, nx+1):
            S[:, [k]] = xbar + (PCoeff)**0.5 * L[:, [k-1]]
        for k in range(nx+1, 2*nx+1):
            S[:, [k]] = xbar - (PCoeff)**0.5 * L[:, [k-nx-1]]

        return S


    assert isinstance(nX, int), "nX must be integer."
    assert isinstance(nU, int), "nU must be integer."
    assert isinstance(nY, int), "nY must be integer."

    assert Qw.shape[0] == Qw.shape[1], "Qw is not square matrix."
    assert Rv.shape[0] == Rv.shape[1], "Rv is not square matrix."

    assert nX == Qw.shape[0], "Shape mismatch of Qw with nX."

    assert nY == Rv.shape[0], "Shape mismatch of Rv with nY."

    assert isinstance(Ts, float), "Sample time (Ts) must be float."
    

    if PCoeff is None or Wm is None or Wc is None:
        L = nX

        PCoeff = alpha**2*(L+kappa)

        W0m = 1-L/(alpha**2*(L+kappa))
        Wm = [W0m] + [1/(2*alpha**2*(L+kappa)) for i in range(2*nX)]

        W0c = 2-alpha**2+beta-L/(alpha**2*(L+kappa))
        Wc = [W0c] + [1/(2*alpha**2*(L+kappa)) for i in range(2*nX)]    

    xp = ca.SX.sym('xp', nX, 1)
    u = ca.SX.sym('u', nU, 1)
    y = ca.SX.sym('y', nY, 1)

    Pp = ca.SX.sym('Pp', nX*nX, 1)
    Ppt = ca.reshape(Pp, nX, nX)
    Sx = GenSigma(xp, Ppt, PCoeff)

    Sxkm = ca.GenSX_zeros(Sx.shape)

    for i in range(2*nX+1):
        Sxkm[:, i] = Integrate(F, Integrator, Ts, Sx[:, i], u)

    mukm = SigmaMean(Sxkm, Wm)
    Pkm = SigmaCovar(Sxkm, Sxkm, Wm, Wc) + Qw

    Sy = ca.GenSX_zeros(nY, 2*nX+1)

    for i in range(2*nX+1):
        Sy[:, i] = H(Sxkm[:, i], u)

    ykm = SigmaMean(Sy, Wm)
    Sk = SigmaCovar(Sy, Sy, Wm, Wc) + Rv

    Ck = SigmaCovar(Sxkm, Sy, Wm, Wc)

    Kk = Ck @ ca.inv(Sk)

    mu = mukm + Kk @ (y - ykm)

    Pk = Pkm - Kk @ Sk @ Kk.T

    return [u, y, xp, Pp], [mu, ca.reshape(Pk, nX*nX, 1)], ['u', 'y', 'xp', 'Pp'], ['xhat', 'Phat']


def simpleMHE(nX, nU, nY, nP, N, Fc, Hc, Wp, Wm, Ts, pLower=[], pUpper=[], arrival=False, GGN=False, Integrator='rk4', Options=None):
    """
    Function to generate simple MHE code using `qrqp` solver. For use with other advanced solver, see `MHE` class.

    Parameters
    ----------
    nX: (int)
        Number of state variables.
    nU: (int)
        number of control input.
    nY: (int)
        Number of measurement variables.
    nP: (int)
        Number of parameter to be estimated. nP=0 while performing state estimation only.
    N: (int)
        Horizon length.
    Fc: (function)
        Function that returns right hand side of state equation.
    Hc: (function)
        Function that returns right hand side of measurement equation.
    Wp: (float or casadi.SX array or numpy.2darray)
        Weight for process noise term. It is $Q_w^{-1/2}$ where $Q_w$ is process noise covariance.
    Wm: (float or casadi.SX array or numpy.2darray)
        Weight for measurement noise term. It is $R_v^{-1/2}$ where $R_v$ is measurement noise covariance.
    Ts: (float): Sample time for MHE
    pLower: (list, optional)
        List of lower limits of unknown parameters. Defaults to [].
    pUpper: (list, optional)
        List of upper limits of unknown parameters. Defaults to [].
    arrival: (bool, optional
         Whether to include arrival cost. Defaults to False.
    GGN: (bool, optional)
        Whether to use GGN. Use this option only when optimization problem is nonlinear. Defaults to False.
    Integrator: (str, optional)
        Integration method. See `BasicUtils.Integrate()` function. Defaults to 'rk4'.
    Options: (dict, optional)
        Option for `qrqp` solver. Defaults to None.

    Returns
    -------
    tuple:
        Tuple of Input, Output, Input name and Output name. Input and output are list of casadi symbolics (`casadi.SX`).
            Input should be control input and measurement data of past horizon length
            Output are all value of decision variable, estimations of parameter, estimates of states and cost function.
    """

    estParam = nP > 0

    X = ca.SX.sym('X', nX, N+1)
    U = ca.SX.sym('U', nU, N)
    Y = ca.SX.sym('Y', nY, N+1)

    J = ca.vertcat()

    # Symbolics for parameters
    if estParam:
        P = ca.SX.sym('P', nP, 1)

    # Arrival cost in cost function formulation
    if arrival:
        xb = ca.SX.sym('xb', nX, 1)
        Wax = ca.SX.sym('Wax', nX*nX, 1)
        J = ca.vertcat(
            J,
            Wax.reshape((nX, nX)) @ (X[:, 0] - xb)
        )
        # Add parameter as well in arrival cost
        if estParam:
            pb = ca.SX.sym('pb', nP, 1)
            Wap = ca.SX.sym('Wap', nP*nP, 1)
            J = ca.vertcat(
                J,
                Wap.reshape((nP, nP)) @ (P - pb)
            )

    # Stage terms in cost function
    for k in range(N):
        J = ca.vertcat(
            J,
            Wm @ (Y[:, k] - Hc(X[:, k])),
            Wp @ (X[:, k+1] - Integrate(Fc, Integrator, Ts, X[:, k], U[:, k], P))
        )

    J = ca.vertcat(
        J,
        Wm @ (Y[:, N] - Hc(X[:, N]))
    )

    # Constraints on parameters
    if estParam:
        g = ca.vertcat()
        lbg = ca.vertcat()
        ubg = ca.vertcat()

        for k in range(nP):
            g = ca.vertcat(
                g,
                P[k]
            )
            lbg = ca.vertcat(
                lbg,
                pLower[k]
            )
            ubg = ca.vertcat(
                ubg,
                pUpper[k]
            )
    else:
        g = None
        lbg = None
        ubg = None

    # decision variables
    if estParam:
        z = ca.vertcat(
            P,
            X.reshape((-1, 1))
        )
    else:
        z = X.reshape((-1, 1))

    pIn = ca.vertcat(
        U.reshape((-1, 1)),
        Y.reshape((-1, 1))
    )

    if arrival:
        pIn = ca.vertcat(
            pIn,
            xb,
            Wax
        )

        if estParam:
            pIn = ca.vertcat(
                pIn,
                pb,
                Wap
            )

    if GGN:
        nlp = nlp2GGN(z, J, g, lbg, ubg, pIn)
        nlp['p'] = ca.vertcat(
            nlp['p'],
            nlp['zOp']
        )
    else:
        nlp = {
            'x': z,
            'f': ca.norm_2(J)**2,
            'g': g,
            'lbg': lbg,
            'ubg': ubg,
            'p': pIn
        }

    optTemp = {'qpsol': 'qrqp'}
    if Options is not None:
        Options.update(optTemp)
    else:
        Options = optTemp

    MHE_prob = {
        'x': nlp['x'],
        'f': nlp['f'],
        'g': nlp['g'],
        'p': nlp['p']
    }

    S = ca.nlpsol('S', 'sqpmethod', MHE_prob, Options)

    xGuess = ca.MX.sym('xGuess', MHE_prob['x'].shape)
    Up = ca.MX.sym('Up', nU, N)
    Yp = ca.MX.sym('Yp', nY, N+1)

    pVal = ca.vertcat(
        Up.reshape((-1, 1)),
        Yp.reshape((-1, 1))
    )

    if arrival:
        xbp = ca.MX.sym('xbp', xb.shape)
        Waxp = ca.MX.sym('Waxp', nX*nX, 1)
        pVal = ca.vertcat(
            pVal,
            xbp,
            Waxp
        )
        if estParam:
            pbp = ca.MX.sym('pbp', nP, 1)
            Wapp = ca.MX.sym('Wapp', nP*nP, 1)
            pVal = ca.vertcat(
                pVal,
                pbp,
                Wapp
            )

    if GGN:
        zOpp = ca.MX.sym('zOpp', z.shape)
        pVal = ca.vertcat(
            pVal,
            zOpp
        )

    if estParam:
        r = S(x0=xGuess, p=pVal, lbg=lbg, ubg=ubg)
    else:
        r = S(x0=xGuess, p=pVal)

    In = [
        xGuess,
        Up,
        Yp,
    ]
    InName = [
        'zGuess',
        'um',
        'ym',
    ]

    if arrival:
        In += [xbp, Waxp]
        InName += ['xL', 'WxL']
        if estParam:
            In += [pbp, Wapp]
            InName += ['pL', 'WpL']

    if GGN:
        In.append(zOpp)
        InName.append('zOp')

    Out = [r['x']]
    OutName = ['zOut']

    if estParam:
        Out.append(r['x'][0:nP])
        OutName.append('p_hat')

    Out.append(r['x'][nP+N*nX:nP+(N+1)*nX])
    OutName.append('x_hat')

    Out.append(r['f'])
    OutName.append('Cost')

    return In, Out, InName, OutName



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyNLControl",
    version="0.0.16",
    author="Niranjan Bhujel",
    author_email="niranjan.bhujel2014@gmail.com",
    description="Package for non-linear control and estimation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NiranjanBhujel/pyNLControl",
    project_urls={
        "Bug Tracker": "https://github.com/NiranjanBhujel/pyNLControl/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    package_data={
        "": ["templates/*.j2", "external/*.zip"],
    },
    python_requires=">=3.6",
    install_requires=[
        "casadi",
        "jinja2",
    ]
)



# Filename:     QPInterface.py
# Written by:   Niranjan Bhujel
# Description:  Contains interface to qp solver. Currently, only qpOASES is supported.


from pynlcontrol import BasicUtils
import casadi as ca
import numpy
import os
from jinja2 import Environment, FileSystemLoader
import zipfile


class qpOASES:
    def __init__(self, H, h, p=None, A=None, lbA=None, ubA=None, lbx=None, ubx=None) -> None:
        """
        Class to create interface to qpOASES solver.

        It can be used to generate to C/C++ code to solve given quadratic optimization problem.

        Parameters
        ----------
            H: (casadi.SX)
                Hessian matrix of cost function
            h: (casadi.SX)
                Linear coefficient vector of cost function
            p: (list, optional)
                List of input parameters to optimization problem. Defaults to None.
            A: (casadi.SX, optional)
                Constraint matrix. Defaults to None.
            lbA: (casadi.SX, optional)
                Lower bound on constraint matrix. Defaults to None.
            ubA: (casadi.SX, optional)
                Upper bound on constraint matrix. Defaults to None.
            lbx: (casadi.SX, optional)
                Lower bound on decision variables. Defaults to None.
            ubx: (casadi.SX, optional)
                Upper bound on decision variables. Defaults to None.
        Returns
        -------
        None


        Example
        -------
        Consider optimization problem:
            min (x1-a)^2 + (x2-b)^2
                s.t. 2x1+3x2 <= 3
                     0 <= x1-x2 <= 10
        
        Solver for this optimization problem can be generated as
        >>> import casadi as ca
        >>> from pynlcontrol import QPInterface
        >>> x1 = ca.SX.sym('x1')
        >>> x2 = ca.SX.sym('x2')
        >>> a = ca.SX.sym('a')
        >>> b = ca.SX.sym('b')
        >>> J = (x1-a)**2 + (x2-b)**2
        >>> H, h, _ = ca.quadratic_coeff(J, ca.vertcat(x1, x2))
        >>> g = ca.vertcat(2*x1+3*x2, x1-x2), ca.vertcat(x1, x2)
        >>> lbg = ca.vertcat(-ca.inf, 0)
        >>> ubg = ca.vertcat(3, 10)
        >>> A, c = ca.linear_coeff(ca.vertcat(2*x1+3*x2, x1-x2), ca.vertcat(x1, x2))
        >>> lbA = lbg - c
        >>> ubA = ubg - c
        >>> qp = QPInterface.qpOASES(H, h, A=A, lbA=lbA, ubA=ubA, p=[a, b])
        >>> qp.exportCode('test1', dir='Test1_Exported', printsfun=True, mex=False, TestCode=False, Options={'max_iter': 5})
        (xGuess[2x1], a[1x1], b[1x1])->(xOpt_VAL[2x1], Obj_VAL[1x1])
        Test1_Exported/test1.c
        Test1_Exported/test1_EVAL_CODE.c
        Test1_Exported/test1_EVAL_CODE_Call.c
        Test1_Exported/BLASReplacement.cpp
        Test1_Exported/Bounds.cpp
        Test1_Exported/Constraints.cpp
        Test1_Exported/Flipper.cpp
        Test1_Exported/Indexlist.cpp
        Test1_Exported/Matrices.cpp
        Test1_Exported/MessageHandling.cpp
        Test1_Exported/Options.cpp
        Test1_Exported/OQPinterface.cpp
        Test1_Exported/QProblem.cpp
        Test1_Exported/QProblemB.cpp
        Test1_Exported/SolutionAnalysis.cpp
        Test1_Exported/SparseSolver.cpp
        Test1_Exported/SQProblem.cpp
        Test1_Exported/SQProblemSchur.cpp
        Test1_Exported/SubjectTo.cpp
        Test1_Exported/Utils.cpp
        #include "Test1_Exported/test1.h"
        #include "Test1_Exported/test1_EVAL_CODE.h"
        #include "Test1_Exported/test1_EVAL_CODE_Call.h"
        // Try this first:
        test1_Call(xGuess, a, b, xOpt_VAL, Obj_VAL);
        // If previous does not work, try this:
        double *xGuess_TEMP = (double *)xGuess;
        double *a_TEMP = (double *)a;
        double *b_TEMP = (double *)b;
        test1_Call(xGuess_TEMP, a_TEMP, b_TEMP, xOpt_VAL, Obj_VAL);


        Running above code generates C/C++ code that can be integrated into other platform. Simulink example is given. Printed output gives name and location of C/C++ files, header files and function call.

        """
        self.H = ca.densify(H)
        self.h = h

        self.OutName = ["H", "g"]
        self.Out = [self.H.T, self.h]
        self.OutSize = [
            f"{self.H.shape[0]}*{self.H.shape[1]}",
            f"{self.h.shape[0]}"
        ]

        self.In = []
        self.InName = []

        if p is not None:
            self.p = p
            self.In += self.p
            self.InName += BasicUtils.__SXname__(self.p)

        if A is not None:
            self.A = ca.densify(A)
            self.OutName.append("A")
            self.Out.append(self.A.T)
            self.OutSize.append(f"{self.A.shape[0]}*{self.A.shape[1]}")

            self.lbA = lbA if lbA is not None else - \
                ca.inf*numpy.ones((self.A.shape[0]))
            self.OutName.append("lbA")
            self.Out.append(self.lbA)
            self.OutSize.append(f"{self.lbA.shape[0]}")

            self.ubA = ubA if ubA is not None else ca.inf * \
                numpy.ones((self.A.shape[0]))
            self.OutName.append("ubA")
            self.Out.append(self.ubA)
            self.OutSize.append(f"{self.ubA.shape[0]}")

        self.lbx = lbx if lbx is not None else -ca.inf*numpy.ones((H.shape[0]))
        self.OutName.append("lbx")
        self.Out.append(self.lbx)
        self.OutSize.append(f"{self.lbx.shape[0]}")

        self.ubx = ubx if ubx is not None else ca.inf*numpy.ones((H.shape[0]))
        self.OutName.append("ubx")
        self.Out.append(self.ubx)
        self.OutSize.append(f"{self.lbx.shape[0]}")

        self.nV = H.shape[0]
        if A is not None:
            self.nC = lbA.shape[0]
        else:
            self.nC = 0

    def exportEvalCode(self, funcname, dir='/', options=None):
        """
        Method of class qpOASES to generate C code to evaluate H, h, A, lbA, ubA, lbx, ubx.

        Parameters
        ----------
            funcname: (str)
                Function to be named that evaluates H, h, A, lbA, ubA, lbx and ubx.
            dir: (str, optional)
                Directory where codes are to be exported. Defaults to '/'.
            options: (dict, optional)
                Options for code generation. Same option as casadi.Function.generate() function. Defaults to None.
        
        Returns
        -------
        None
        
        """
        if options is None:
            options = {
                'main': False,
                'with_header': True
            }
        
        Func = ca.Function(
            funcname,
            self.In,
            self.Out,
            self.InName,
            self.OutName
        )

        BasicUtils.Gen_Code(Func, funcname+'_CODE', dir, printhelp=False)

    def exportCode(self, filename, dir='/', mex=False, printsfun=False, Options=None, TestCode=False):
        """
        Method of class qpOASES to export the code that solves quadratic programming.

        Parameters
        ----------
            filename: (str)
                Filename for exported code.
            dir: (str, optional)
                Directory where code is to be exported. Defaults to '/'.
            mex: (bool, optional)
                Whether mex interface is required. Defaults to False.
            printsfun: (bool, optional)
                Whether MATLAB s-function interface is to be implemented. Defaults to False.
            Options: (dict, optional)
                Options for qpOASES solver. Defaults to None.
            TestCode: (bool, optional)
                Whether main function is required. Useful while testing and debugging. Defaults to False.
        
        Returns
        -------
        None
        """

        if Options is not None:
            if "max_iter" in Options:
                MaxIter = Options['max_iter']
            else:
                MaxIter = 10
            Options.pop('max_iter', None)
            Opt = Options
        else:
            Opt = {}
            MaxIter = 10

        if dir[-1] != '/':
            tmpdir = dir + '/'
        else:
            tmpdir = dir[0:-1]

        with zipfile.ZipFile(f"{os.path.dirname(os.path.realpath(__file__))}\\external\\qpOASES.zip", 'r') as zip_ref:
            zip_ref.extractall(tmpdir)

        self.exportEvalCode(filename + "_EVAL", dir)

        OutList = [
            'A',
            'lbx',
            'ubx',
            'lbA',
            'ubA'
        ]
        OutPointer = []
        for tmp in OutList:
            if tmp in self.OutName:
                OutPointer.append(tmp)
            else:
                OutPointer.append("NULL")

        # template rendering
        templatePath = f"{os.path.dirname(os.path.realpath(__file__))}/templates"

        file_loader = FileSystemLoader(templatePath)
        env = Environment(loader=file_loader)

        # template for c file
        template = env.get_template('qpOASES_c.j2')

        CodeOutput_c = template.render(
            FileName=filename,
            InName=self.InName,
            OutName=self.OutName,
            OutSize=self.OutSize,
            nV=self.nV,
            nC=self.nC,
            OutPointer=OutPointer,
            Options=Opt,
            MaxIter=MaxIter,
            Mex=mex
        )

        template = env.get_template('qpOASES_h.j2')

        CodeOutput_h = template.render(
            FileName=filename,
            InName=self.InName,
            OutName=self.OutName,
            OutSize=self.OutSize,
            nV=self.nV,
            nC=self.nC,
            OutPointer=OutPointer,
            Options=Opt,
            MaxIter=MaxIter,
            Mex=mex
        )

        if mex:
            template = env.get_template('Gen_Mex_c.j2')

            MexOutput_c = template.render(
                InNum=len(self.In)+1,
                InName=["xGuess"] + self.InName,
                OutNum=2,
                OutName=["xOpt", "Obj_VAL"],
                OutSize=[self.H.shape[0], 1],
                FuncCall=f"{filename}_Call"
            )

            CodeOutput_c += "\n\n" + MexOutput_c

        with open(f"{tmpdir}{filename}.c", "w") as fw:
            fw.write(CodeOutput_c)

        with open(f"{tmpdir}{filename}.h", "w") as fw:
            fw.write(CodeOutput_h)

        if mex == True:
            Fin = [f'xGuess[{self.H.shape[0]}x1]']
            for k in range(len(self.InName)):
                Fin.append(
                    str(self.InName[k]) + f'[{self.In[k].shape[0]}x{self.In[k].shape[1]}]')
            Fin = ', '.join(Fin)
            Fout = [f'xOpt_VAL[{self.H.shape[0]}x{1}]', 'Obj_VAL[1x1]']
            Fout = ', '.join(Fout)

            # Print inputs and outputs with their respective size
            print(f'\n({Fin})->({Fout})\n')
        if printsfun and mex == False:
            Fin = [f'xGuess[{self.H.shape[0]}x1]']
            for k in range(len(self.InName)):
                Fin.append(
                    str(self.InName[k]) + f'[{self.In[k].shape[0]}x{self.In[k].shape[1]}]')
            Fin = ', '.join(Fin)
            Fout = [f'xOpt_VAL[{self.H.shape[0]}x{1}]', 'Obj_VAL[1x1]']
            Fout = ', '.join(Fout)

            # Print inputs and outputs with their respective size
            print(f'\n({Fin})->({Fout})\n')

            FuncIn_Str = ', '.join(self.InName)
            # print('INC_PATH qpOASES/include')
            print(f'{tmpdir}{filename}.c')
            print(f'{tmpdir}{filename}_EVAL_CODE.c')
            print(f'{tmpdir}{filename}_EVAL_CODE_Call.c')
            print(f"{tmpdir}BLASReplacement.cpp")
            print(f"{tmpdir}Bounds.cpp")
            print(f"{tmpdir}Constraints.cpp")
            print(f"{tmpdir}Flipper.cpp")
            print(f"{tmpdir}Indexlist.cpp")
            print(f"{tmpdir}LAPACKReplacement.cpp")
            print(f"{tmpdir}Matrices.cpp")
            print(f"{tmpdir}MessageHandling.cpp")
            print(f"{tmpdir}Options.cpp")
            print(f"{tmpdir}OQPinterface.cpp")
            print(f"{tmpdir}QProblem.cpp")
            print(f"{tmpdir}QProblemB.cpp")
            print(f"{tmpdir}SolutionAnalysis.cpp")
            print(f"{tmpdir}SparseSolver.cpp")
            print(f"{tmpdir}SQProblem.cpp")
            print(f"{tmpdir}SQProblemSchur.cpp")
            print(f"{tmpdir}SubjectTo.cpp")
            print(f"{tmpdir}Utils.cpp")

            print('\n')
            print(f'#include "{tmpdir}{filename}.h"')
            print(f'#include "{tmpdir}{filename}_EVAL_CODE.h"')
            print(f'#include "{tmpdir}{filename}_EVAL_CODE_Call.h"\n')

            print('\n// Try this first:')
            print(f'{filename}_Call(xGuess, {FuncIn_Str}, xOpt_VAL, Obj_VAL);')
            print('\n// If previous does not work, try this:')
            print(f"double *xGuess_TEMP = (double *)xGuess;")
            for k in range(len(self.InName)):
                print(
                    f"double *{self.InName[k]}_TEMP = (double *){self.InName[k]};")

            FuncIn = []
            for tmp in self.InName:
                FuncIn.append(str(tmp) + '_TEMP')
            FuncIn_Str = ', '.join(FuncIn)
            print(f'{filename}_Call(xGuess_TEMP, {FuncIn_Str}, xOpt_VAL, Obj_VAL);\n')

        if TestCode:
            Headers = [f"{filename}.h", f"{filename}_EVAL_CODE.h",
                       f"{filename}_EVAL_CODE_Call.h"]
            VarsIn = ['xGuess'] + self.InName
            SizeIn = [self.H.shape[0]] + [self.In[k].shape[0] *
                                          self.In[k].shape[1] for k in range(len(self.In))]
            VarsOut = ['xOpt_VAL', 'Obj_VAL']
            SizeOut = [self.H.shape[0], 1]
            Call_FuncName = f'{filename}_Call'

            BasicUtils.Gen_Test(
                Headers,
                VarsIn,
                SizeIn,
                VarsOut,
                SizeOut,
                Call_FuncName,
                'Main_'+filename+'.c',
                dir
            )
